#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <generate_err.h>
#include <hdr_parser.h>
#include <logging.h>
#include <buffers.h>
#include <hash.h>
#include <const.h>
#include <validate.h>
#include <struct.h>
#include <urlparse.h>

int must_cache_request(char *request)
{
	int type;
	char *tmp;

	type = get_request_type(request);

	if ( (type == REQ_TRACE) || (type == REQ_OPTIONS) )
		return 0;

	if ((tmp=parse_http_hdr("Cache-Control",request)) != NULL)
	{
		if (cache_response_control(tmp) == NO_CACHE)
		{
			free(tmp);
			return 0;
		}
		free(tmp);
	}

	return 1;
}

int must_cache_reply(char *reply)
{
	int type;
	char *tmp=NULL;

	type = get_reply_type(reply);

	if ( (type==206) || (type==303) || (type==304) ||
	     ((tmp=parse_http_hdr("Content-Range",reply)) != NULL) )
	{
		free(tmp);
		return 0;
	}
	free(tmp);

	if ((tmp=parse_http_hdr("Cache-Control",reply)) != NULL)
	{
		if (cache_response_control(tmp) == NO_CACHE)
		{
			free(tmp);
			return 0;
		}
		free(tmp);
	}

	return 1;
}

int must_read_from_cache(char *request)
{
	char *tmp;

	// We jump over the cache (for reading, not for writing)
	// for condiional requests
	if ((tmp=parse_http_hdr("If-Modified-Since",request)) != NULL)
	{
		free(tmp);
		return 0;
	}
	if ((tmp=parse_http_hdr("If-Unmodified-Since",request)) != NULL)
	{
		free(tmp);
		return 0;
	}
	if ((tmp=parse_http_hdr("If-Match",request)) != NULL)
	{
		free(tmp);
		return 0;
	}
	if ((tmp=parse_http_hdr("If-None-Match",request)) != NULL)
	{
		free(tmp);
		return 0;
	}

	// Ignore cache also if Cache-Control contains max-age or min-fresh
	if ((tmp=parse_http_hdr("Cache-Control",request)) != NULL)
	{
		if (strstrcase(tmp,"max-age") != NULL)
		{
			free(tmp);
			return 0;
		}
		if (strstrcase(tmp,"min-fresh") != NULL)
		{
			free(tmp);
			return 0;
		}
		free(tmp);
	}

	return (must_cache_request(request));
}

int must_write_to_cache(char *request,char *reply)
{
	// request may be NULL
	if (request != NULL)
	{
		return ( (must_cache_request(request)) && (must_cache_reply(reply)) );
	}
	else
	{
		return (must_cache_reply(reply));
	}
}

struct reply get_from_cache(char *request, char *host, int *port,
                            char **path, int *connection)
{
	struct reply ret;

	if ( (request == NULL) || (host == NULL) || (port == NULL) ||
	     (path == NULL) )
	{
		log(LOG_ERR,"get_from_cache: called with invalid parameter");
		ret.data=NULL;
		ret.len=-1;
		return ret;
	}

	if (get_request_type(request) == REQ_UNKNOWN)
	{
		char *url;
		log(LOG_WARNING,"get_from_cache: unknown request type");
		url=get_requested_url(request);
		ret=error_reply((url==NULL)?"unknown":url,500);
		*connection=0;
		free(url);
		return ret;
	}

	if ( (*port = get_requested_port(request)) == 0 )
	{
		char *url;
		log(LOG_WARNING,"get_from_cache: invalid request");
		url=get_requested_url(request);
		ret=error_reply((url==NULL)?"unknown":url,400);
		*connection=0;
		free(url);
		return ret;
	}
	if ( (*path = get_requested_path(request)) == NULL )
	{
		char *url;
		log(LOG_WARNING,"get_from_cache: invalid request");
		url=get_requested_url(request);
		ret=error_reply((url==NULL)?"unknown":url,400);
		*connection=0;
		free(url);
		return ret;
	}

	if ( (strstrcase(request,"HTTP/1.0\r\n")) != NULL )
	{
		char *url;
		log(LOG_WARNING,"Browser speaks HTTP/1.0");
		url=get_requested_url(request);
		ret=error_reply((url==NULL)?"unknown":url,505);
		*connection=0;
		free(url);
		return ret;
	}

	if (must_read_from_cache(request))
	{
		struct document *doc;
		struct urlinfo_t url;

		strncpy(url.resolved_host,host,MAX_RESOLVEDHOST_LEN);
		url.resolved_host[MAX_RESOLVEDHOST_LEN - 1]='\0';
		strncpy(url.path,*path,MAX_PATH_LEN);
		url.path[MAX_PATH_LEN - 1]='\0';
		url.port=*port;
		isstream(&url);

		// Retrieve from cache
		if ( (doc=get_doc(&url)) == NULL )
		{
			char *url;
			log(LOG_ERR,"get_from_cache: Cache error");
			url=get_requested_url(request);
			ret=error_reply((url==NULL)?"unknown":url,503);
			*connection=0;
			free(url);
			return ret;
		}
		else if ( doc == (struct document *)-1 )
		{
			ret.data=NULL;
			ret.len=-1;
		}
		else
		{
			if ( (ret.data = malloc(strlen(doc->hdr)+doc->doclen+1)) == NULL )
			{
				ret.len=-1;
			}
			else
			{
				strcpy(ret.data,doc->hdr);
				memcpy(&ret.data[strlen(ret.data)],doc->doc,doc->doclen);
				ret.len = strlen(doc->hdr)+doc->doclen;
				ret.valid = doc->valid;
			}
			free(doc->doc);
			free(doc->hdr);
			free(doc);
		}
	}
	else
	{
		ret.data=NULL;
		ret.len=-1;
	}

	return ret;
}

