#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <wait.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef NONBLOCKING_RW

#include <fcntl.h>

#endif

#include <sock_utils.h>
#include <logging.h>
#include <init_shared.h>
#include <generate_err.h>
#include <rw.h>
#include <hdr_parser.h>
#include <hdr_modify.h>
#include <cache.h>
#include <buffers.h>
#include <get_url.h>
#include <read_url.h>
#include <hash.h>
#include <validate.h>
#include <prefetch.h>
#include <struct.h>
#include <tree.h>

#define LISTEN_PORT 8080

int start_prefetch_lev;
int stop_prefetch;
pid_t prefetching_son;

// This function is called with a SIG_USR1 to stop a prefetch
// process before starting a new one.
void sig_usr1_prefetch(int sig_no)
{
	signal(SIGUSR1,sig_usr1_prefetch);
	stop_prefetch=1;
}

// This function is attached to SIG_CHLD (child died) signal.
// Calling wait is needed to avoid leaving zombie processes around.
void sig_child(int sig_no)
{
	int st;
	signal(SIGCHLD,sig_child);
	prefetching_son = 0;
	wait(&st);
	if (WIFSIGNALED(st))
		log(LOG_DEBUG,"child killed by signal %d\n",WTERMSIG(st));
}

// This function free all shared data structures. It is attached t
// SIG_INT and SIG_TERM signals.
void sig_int(int sig_no)
{
	// free shared memory
	finish_hash();
	detach_shmem();
	drop_shmem();

	exit(0);
}

// All spawned process must call this function when starting,
// because only one process must free shared data structures 
// when exiting
void drop_signals()
{
	// No error check here. If something goes wrong there will be
	// just some error log when exiting, but it's fine.
	if (signal(SIGINT,SIG_DFL) == SIG_ERR)
	{
		log(LOG_ERR,"Cannot detach signal handler");
	}
	if (signal(SIGTERM,SIG_DFL) == SIG_ERR)
	{
		log(LOG_ERR,"Cannot detach signal handler");
	}
}

// This function reads a request from the browser, and attached data
// (if present).
// If NONBLOCKING_RW is defined, return value can be (char *)-1 to indicate
// that the connection has timed out
char *read_request(int fd)
{
	// Here we assume that 4096 bytes will be able to contain every request.
	// Longer request will be cutted at 4096 bytes.
	char buf[4096];
	int nread=0;
	char *ptr;
	char *len;
	char *ret;

	do
	{
		int nr;
		if ( (nr = _read(fd,&buf[nread],(sizeof(buf)-nread)-1)) <= 0 )
		{
#ifdef NONBLOCKING_RW
			if (nr == -EAGAIN)
				return (char*)-1;
#endif
			log(LOG_WARNING,"read_request: read error "
			                "(client closed connection?)");
			return NULL;
		}

		nread += nr;
		buf[nread]='\0';
	}
	while ( ((ptr=get_http_doc_start(buf)) == NULL) && (nread < sizeof(buf)) );

	if ( (ret=malloc(nread+1)) == NULL )
	{
		log(LOG_WARNING,"read_request: not enough memory");
		return NULL;
	}
	memcpy(ret,buf,nread);
	buf[nread]='\0';

	// Read extra data (if present)
	len=parse_http_hdr("Content-Length",buf);

	if (len != NULL)
	{
		char *data;
		int ilen=atoi(len);

		free(len);
		ilen -= (&buf[nread]-ptr);

		if ( (data=malloc(ilen+1)) == NULL )
		{
			log(LOG_WARNING,"read_request: not enough memory");
			free(ret);
			return NULL;
		}

		if (_readn(fd,data,ilen) < ilen)
		{
			free(data);
			free(ret);
			return NULL;
		}
		data[ilen]='\0';

		if (bufcat(&ret,data) == NULL)
		{
			// Ouch!! We ran out of memory...
			log(LOG_WARNING,"read_request: not enough memory");
			free(ret);
			free(data);
			return NULL;
		}
		free(data);
	}

	return ret;
}

// Returns 1 if document is html, 0 otherwise
int must_prefetch(char *header)
{
	char *ptr;

	if ( (ptr=parse_http_hdr("Content-type",header)) == NULL )
	{
		return 0;
	}

	if (strstrcase(ptr,"text/html") != NULL)
	{
		free(ptr);
		return 1;
	}
	else
	{
		free(ptr);
		return 0;
	}
}

void start_prefetch(struct urlinfo_t url)
{
	int cache_found;
	char *host_resolved;
	struct document *doc_tmp;
	struct document doc;

	if (start_prefetch_lev < 0)
	{
		start_prefetch_lev = url.lev-1;
	}

	if (url.resolved_host[0] == '\0')
	{
		if ( (host_resolved=resolve_host(url.host)) != NULL )
		{
			strncpy(url.resolved_host,host_resolved,MAX_RESOLVEDHOST_LEN);
			url.resolved_host[MAX_RESOLVEDHOST_LEN-1]='\0';
			free(host_resolved);
		}
		else
		{
			// Use host not resolved if cannot resolve
			strncpy(url.resolved_host,url.host,MAX_RESOLVEDHOST_LEN);
		}
	}

	if ( (doc_tmp=get_doc(&url)) == (struct document *)-1 )
	{
#ifdef LOG_TO_SCREEN
		log(LOG_DEBUG,"prefetching %s %d %s",url.host,url.port,url.path);
#endif
		cache_found=0;
		if ( (doc=get_document(url.host,
		           (url.resolved_host[0]=='\0') ? NULL : url.resolved_host,
		           url.port,url.path,NULL,-1,5)).hdr == NULL )
		{
			// If doc.hdr is null there have been problems with the SERVER.
			log(LOG_WARNING,"start_prefetch: cannot retrieve document %s%s",
			    url.host,url.path);

			// The cache is still waitin' for the document
			clear_one(&url);
			return;
		}
		url.size = doc.doclen;
	}
	else if (doc_tmp == NULL)
	{
		log(LOG_ERR,"start_prefetch: Cache error");
		return;
	}
	else
	{
		cache_found = 1;
		doc = *doc_tmp;
		free(doc_tmp);
	}

	if ( doc.hdr == NULL )
	{
		// The cache is still waitin' for the document
		if (!cache_found) clear_one(&url);
	}
	else if (doc.doc != NULL)
	{
		if (must_write_to_cache(NULL,doc.hdr))
		{
			if (!cache_found)
			{
				// Insert into cache
				if (insert_file(&url,&doc) != 1)
				{
					log(LOG_ERR,"start_prefetch: insert_file failed");
					// Failed before making the V
				}
				clear_one(&url);
			}

			if (must_prefetch(doc.hdr))
			{
				bufncat(&doc.doc,doc.doclen,"\0",1);
				url.prefetch = 1; // page is requested for prefething
				searchforhref (doc.doc,url); // continue prefetch
				if (url.lev - start_prefetch_lev < MAX_PREFETCH_LEVEL)
					send_req(PPP); // create queue for prefetch
			}
		}
		else
		{
			// The cache is still waitin' for the document
			if (!cache_found) clear_one(&url);
		}
		free(doc.doc);
		free(doc.hdr);
	}
	else
	{
		// The cache is still waitin' for the document
		if (!cache_found) clear_one(&url);
		free(doc.hdr);
	}

	if (stop_prefetch)
	{
		detach_shmem();
		exit(0);
	}

	return;
}

// Transform a struct reply into a struct document, then start prefetching.
// If must_fork = 1 it goes into background for prefetching.
void prefetch_page(struct reply rep, char *host, char *resolved_host, int port,
                   char *path, int must_fork)
{
	struct document doc;
	struct urlinfo_t url;
	char *ptr;
	char backup;

	if (!must_prefetch(rep.data))
	{
		return;
	}
	if (must_fork)
	{
		int f_rv;

		// Stop current prefetching process
		if (prefetching_son != 0)
		{
			// No error check because child could have ended,
			// so even an error can be ok.
			kill(prefetching_son,SIGUSR1);
			prefetching_son=0;
		}

		f_rv = fork();

		if (f_rv < 0)
		{
			log(LOG_ERR,"prefetch_page: cannot fork");
			return;
		}
		else if (f_rv > 0)
		{
			prefetching_son = f_rv;
			return;
		}
		drop_signals();
		if (!attach_shmem())
		{
			log(LOG_ERR,"prefetch_page: Cannot attach to shared memory");
			exit(1);
		}
	}

	ptr=get_http_doc_start(rep.data);

	if (ptr == NULL)
	{
		log(LOG_WARNING,"prefetch_page: Cannot find header");
		if (must_fork)
		{
			drop_shmem();
			exit(0);
		}
		else
			return;
	}

	if ( (ptr - rep.data) < rep.len )
	{
		// Writes a temporary '\0' to mark the end of the headers
		backup = *ptr;
		*ptr = '\0';
		if ( (doc.hdr = malloc(strlen(rep.data)+1)) == NULL )
		{
			log(LOG_ERR,"prefetch_page: Not enough memory");
			if (must_fork)
			{
				drop_shmem();
				exit(0);
			}
			else
				return;
		}
	
		strcpy (doc.hdr,rep.data);
		*ptr = backup;

		doc.doclen = rep.len - strlen(doc.hdr);
		if ( (doc.doc = malloc(doc.doclen)) == NULL )
		{
			log(LOG_ERR,"prefetch_page: Not enough memory");
			free(doc.hdr);
			if (must_fork)
			{
				drop_shmem();
				exit(0);
			}
			else
				return;
			}

		memcpy(doc.doc,rep.data + strlen(doc.hdr),doc.doclen);
		bufncat(&doc.doc,doc.doclen,"\0",1);
	}
	else
	{
		if ( (doc.hdr = malloc(rep.len+1)) == NULL )
		{
			log(LOG_ERR,"prefetch_page: Not enough memory");
			if (must_fork)
			{
				drop_shmem();
				exit(0);
			}
			else
				return;
		}

		memcpy (doc.hdr,rep.data,rep.len);
		doc.hdr[rep.len]='\0';

		doc.doc = "";
		doc.doclen = 0;
	}

	strncpy(url.resolved_host,resolved_host,MAX_RESOLVEDHOST_LEN);
	url.resolved_host[MAX_RESOLVEDHOST_LEN - 1]='\0';
	strncpy(url.host,host,MAX_HOST_LEN);
	url.host[MAX_HOST_LEN - 1]='\0';
	strncpy(url.path,path,MAX_PATH_LEN);
	url.path[MAX_PATH_LEN - 1]='\0';
	url.port = port;
	url.size = doc.doclen;

	// Start prefetching
	searchforhref (doc.doc,url);
	start_prefetch_lev = -1;
	send_req(PPP);
	obtain_lead();

	if (doc.doclen > 0) free(doc.doc);
	free(doc.hdr);
	if (must_fork)
	{
		drop_shmem();
		exit(0);
	}
	return;
}

// This function reads a document from a remote server, streaming it to
// the browser, then the spawned process sends the document to the cache,
// starts prefetching and exits
//
// Return values:
//  2 : Cache document is valid (304)
//  1 : Everything ok
//  0 : Could not open connection - can answer with an error page
// -1 : Error during connection - close connection with client immediatly
int get_parse_cache_reply(char *host,char *resolved_host,int port,
                          char *path,int reply_fd,char **request,int validate,
                          struct reply rep)
{
	int p[2];
	int f_rv;

	if (pipe(p) < 0)
	{
		log(LOG_ERR,"get_parse_cache_reply: cannot create pipe");
		return 0;
	}

	if ( (f_rv=fork()) < 0 )
	{
		log(LOG_ERR,"get_parse_cache_reply: cannot fork");
		return 0;
	}

	if (f_rv==0)
	{
		// Son
		struct document doc;
		struct urlinfo_t url;

		drop_signals();
		if (attach_shmem() != 1)
		{
			log(LOG_ERR,"Cannot attach to shared memory");
			exit(1);
		}

		stop_prefetch = 0;
		if (signal(SIGUSR1,sig_usr1_prefetch) == SIG_ERR)
		{
			log(LOG_ERR,"get_parse_cache_reply: Cannot attach signal handler");
			stop_prefetch=1;
		}

		close(reply_fd);
		close(p[0]);

		doc=get_document(host,resolved_host,port,path,request,p[1],10);

		close(p[1]);

		strncpy(url.resolved_host,resolved_host,MAX_RESOLVEDHOST_LEN);
		url.resolved_host[MAX_RESOLVEDHOST_LEN - 1]='\0';
		strncpy(url.host,host,MAX_HOST_LEN);
		url.host[MAX_HOST_LEN - 1]='\0';
		strncpy(url.path,path,MAX_PATH_LEN);
		url.path[MAX_PATH_LEN - 1]='\0';
		url.port = port;
		url.size = doc.doclen;

		if ( doc.hdr == NULL )
		{
			// If doc.hdr is null there have been problems with the SERVER.
			log(LOG_WARNING,"get_parse_cache_reply: cannot retrieve document");

			// The cache is still waitin' for the document
			clear_one(&url);
		}
		else if (doc.doc != NULL)
		{
			if (must_write_to_cache(*request,doc.hdr))
			{
				// Insert into cache
				if ((insert_file(&url,&doc)) == 0)
				{
					// Failed before making the V
					clear_one(&url);
				}
				else if (validate)
				{
					// insert_file function replaced the
					// document and it needs a V
					clear_one(&url);
				}
			}
			else
			{
				// The cache is still waitin' for the document
				clear_one(&url);
			}

			if ((!stop_prefetch) && (must_prefetch(doc.hdr)))
			{
				// Start prefetching
				// No need to call prefetch_page here, because we have
				// the needed stuctures yet, and we must not fork.
				bufncat(&doc.doc,doc.doclen,"\0",1);
				searchforhref (doc.doc,url);
				start_prefetch_lev = -1;
				send_req(PPP);
				obtain_lead();
			}

			free(doc.doc);
			free(doc.hdr);
		}
		else
		{
			// The cache is still waitin' for the document
			clear_one(&url);

			// check if answer is 304 (must prefetch)
			if ( (validate) && (get_reply_type(doc.hdr) == 304) &&
			     (!stop_prefetch) && (rep.data != NULL) )
			{
				prefetch_page(rep,url.host,url.resolved_host,url.port,
				              url.path,0);
			}
			free(doc.hdr);
		}

		// Son dies here
		detach_shmem();
		exit(0);
	}
	else
	{
		// Father
		char buf[4096];
		int nr;
		int tot_read=0;

		close(p[1]);

		// If we are validating the page check if reply is 304.
		if ( (validate) || (prefetching_son != 0) )
		{
			char *headers;
			char *read_ptr=NULL;
			int nw=-1;

			if ( (headers=read_headers(p[0],&read_ptr,&tot_read,&nw))==NULL )
			{
				log(LOG_ERR,"get_parse_cache_reply: error reading headers");
				close(p[0]);
				return 0;
			}

			if ( (prefetching_son != 0) && (must_prefetch(headers)) )
			{
				// Stop current prefetching process
				// No error check because child could have ended,
				// so even an error can be ok.
				kill(prefetching_son,SIGUSR1);
				prefetching_son=f_rv;
			}

			if ( (validate) && ((get_reply_type(headers)) == 304) )
			{
				close(p[0]);
				free(headers);
				free(read_ptr);
				return 2;
			}

			if ( (nw=_writen(reply_fd,headers,strlen(headers)))
			         < strlen(headers))
			{
				close(p[0]);
				free(headers);
				free(read_ptr);
				return (nw==0)?0:-1;
			}
			free(headers);
			if (_writen(reply_fd,read_ptr,tot_read) < tot_read)
			{
				close(p[0]);
				free(read_ptr);
				return -1;
			}
			free(read_ptr);
		}

		while ( (nr=_read(p[0],buf,sizeof(buf))) > 0 )
		{
			tot_read+=nr;
			if (_writen(reply_fd,buf,nr) < nr)
			{
				close(p[0]);
				return -1;
			}
		}
		close(p[0]);

		if (tot_read == 0)
		{
			// Hmmmm... son must have died before sending anything...
			// We can send an "Internal Error" answer to the browser.
			return 0;
		}

		if (nr < 0)
		{
			// If nr here is less than 0 something must have gone wrong in
			// the _read call, so we badly close this connection
			return -1;
		}
	}

	return 1;
}

// This function manage a connection with the browser, reading the requests
// and sending appropriate answers
void manage_connection(int fd)
{
	char *request;
	struct reply rep;
	int connection=1; // this variable is set to 0 when the connection with
	                  // the browser must be closed
	prefetching_son = 0; // this varaible holds the pid of the prefetching
	                     // son.
	do
	{
		request=read_request(fd);

		if (request == NULL)
		{
			rep=error_reply("unknown",503);
			// Checking returned value here is unuseful, because we will
			// close the connection aven if the write goes well
			_writen(fd,rep.data,rep.len);
			connection=0;
			free(rep.data);
		}
#ifdef NONBLOCKING_RW
		else if (request == (char *)-1)
		{
			// Connection with the client timed out... simply close the
			// socket
			connection=0;
		}
#endif
		// Check if the request is valid
		else if ( (rep=immediate_answer(&request)).data != NULL )
		{
			// Checking returned value here is unuseful, because we will
			// close the connection aven if the write goes well
			_writen(fd,rep.data,rep.len);
			connection=0;
			free(rep.data);
		}
		else
		{
			// Valid request
			int port;
			char *tmp;
			char *path;
			char *host;
			char *host_resolved;

			if ( ((tmp=parse_http_hdr("Connection",request)) != NULL) &&
			     (strstrcase(tmp,"close") != NULL) )
			{
				// The client is asking us to close the connection
				connection=0;
			}
			free(tmp);

			// Proxy-Connection is not fully HTTP/1.1 compliant, but some
			// browsers still use that
			if ( ((tmp=parse_http_hdr("Proxy-Connection",request)) != NULL) &&
			     (strstrcase(tmp,"close") != NULL) )
			{
				// The client is asking us to close the connection
				connection=0;
			}
			free(tmp);

			/* Resolve the hostname */
			if ( (host=get_requested_host(request)) == NULL )
			{
				host_resolved=NULL;
			}
			else
			{
				host_resolved=resolve_host(host);
			}

			if (host_resolved==NULL)
			{
				char *url;
				url=get_requested_url(request);
				rep=error_reply((url==NULL)?"unknown":url,503);
				// Checking returned value here is unuseful, because we will
				// close the connection aven if the write goes well
				_writen(fd,rep.data,rep.len);
				connection=0;
				free(rep.data);
				free(host);
				free(url);
			}
			else if ( ( (rep=get_from_cache(request,host_resolved,&port,
			                              &path,&connection)).data != NULL ) &&
			          ( rep.valid == 1 ) )
			{
				// Ok, it's in cache: reply
				if (_writen(fd,rep.data,rep.len) < rep.len)
				{
					// Close the connection
					connection=0;
				}
				prefetch_page(rep,host,host_resolved,port,path,1);
				free(host_resolved);
				free(host);
				free(path);
				free(rep.data);
			}
			else
			{
				int rv;
				int validate;

				if (rep.data == NULL)
				{
					validate = 0;
				}
				else
				{
					validate = !rep.valid;
					// If validate == 1 modify request
					if (validate)
					{
						if (prepare_validation_header(rep.data,&request) < 0)
						{
							char *url;
							struct reply tmp_rep;
							struct urlinfo_t urlinfo;
							log(LOG_ERR,"manage_connection: cannot modify "
							            "request (validation)");
							url=get_requested_url(request);
							tmp_rep=error_reply((url==NULL)?"unknown":url,503);
							// Checking returned value here is unuseful, because
							// we will close the connection aven if the write
							// goes well
							_writen(fd,tmp_rep.data,tmp_rep.len);
							free(url);
							free(tmp_rep.data);
							connection=0;

							// The cache is still waiting for the document
							strncpy(urlinfo.resolved_host,host_resolved,
							        MAX_RESOLVEDHOST_LEN);
							urlinfo.resolved_host[MAX_RESOLVEDHOST_LEN-1]='\0';
							strncpy(urlinfo.host,host,MAX_HOST_LEN);
							urlinfo.host[MAX_HOST_LEN - 1]='\0';
							strncpy(urlinfo.path,path,MAX_PATH_LEN);
							urlinfo.path[MAX_PATH_LEN - 1]='\0';
							urlinfo.port = port;

							clear_one(&urlinfo);
						}
					}
				}

				if (connection)
				{
					// Read from server, write to client, parse for prefetching
					// and caching
					if ( (rv=get_parse_cache_reply(host,host_resolved,port,path,
					                           fd,&request,validate,rep)) < 1)
					{
						if (rv==0)
						{
							char *url;
							struct reply tmp_rep;
							url=get_requested_url(request);
							tmp_rep=error_reply((url==NULL)?"unknown":url,503);
							// Checking returned value here is unuseful, because
							// we will close the connection aven if the write
							// goes well
							_writen(fd,tmp_rep.data,tmp_rep.len);
							free(url);
							free(tmp_rep.data);
						}
						connection=0;
					}
					else if ( rv == 2 )
					{
						// Ok, cache document is valid: reply
						if (_writen(fd,rep.data,rep.len) < rep.len)
						{
							// Close the connection
							connection=0;
						}
					}

				}
				free(rep.data);
				free(host_resolved);
				free(host);
				free(path);
			}
		}
	}
	while (connection);

#if 0
	// Stop current prefetching process when browser closes its connection
	// (this saves bandwidth, but with browsers like mozilla - which closes
	// connection after every transaction - makes prefetching really
	// unuseful)
	if (prefetching_son != 0)
	{
		// No error check because child could have ended,
		// so even an error can be ok.
		kill(prefetching_son,SIGUSR1);
		prefetching_son=0;
	}
#endif
}

// This is the main function of the proxy. It opens the main socket,
// waits for connection and spawn childs to manage communications with
// browsers.
// Return 1 if shared structure must be deallocated, 0 otherwise
int proxy_main()
{
	int rv;
	int connfd,listenfd;

	/* open listening socket */
	if ( (rv=tcp_listen(&listenfd,LISTEN_PORT)) != 0 )
	{
		log(LOG_ERR,"Cannot open listening socket, errno: %d",rv);
		return 0;
	}

	/* Set up signal handlers */
	if (signal(SIGCHLD,sig_child) == SIG_ERR)
	{
		log(LOG_ERR,"Cannot attach signal handler");
		return 0;
	}
	if (signal(SIGPIPE,SIG_IGN) == SIG_ERR)
	{
		log(LOG_ERR,"Cannot attach signal handler");
		return 0;
	}
	if (signal(SIGINT,sig_int) == SIG_ERR)
	{
		log(LOG_ERR,"Cannot attach signal handler");
		return 0;
	}
	if (signal(SIGTERM,sig_int) == SIG_ERR)
	{
		log(LOG_ERR,"Cannot attach signal handler");
		return 0;
	}

	/* setup shared memory */
	if (!init_shared())
		return 0;

	if (!init_hash())
		return 1;

	while (1)
	{

#ifdef NONBLOCKING_RW
		int flags;
#endif

		connfd=accept(listenfd,NULL,NULL);

		// This check has been added after the meeting with Vittorio Ghini
		if (connfd >= 0)
		{
#ifdef NONBLOCKING_RW
			if ( (flags=fcntl(connfd,F_GETFL,0)) < 0 )
			{
				log(LOG_ERR,"Cannot set connfd to O_NONBLOCK");
				close(connfd);
			}
			else
			{
				flags |= O_NONBLOCK;
				if ( (fcntl(connfd,F_SETFL,flags)) < 0 )
				{
					log(LOG_ERR,"Cannot set connfd to O_NONBLOCK");
					close(connfd);
				}
				else
				{
#endif
					if (fork()==0)
					{
						close(listenfd);
						drop_signals();
						if (attach_shmem() != 1)
						{
							log(LOG_ERR,"Cannot attach to shared memory");
							exit(1);
						}
						manage_connection(connfd);
						close(connfd);
						detach_shmem();
						exit(0);
					}
					close(connfd);
#ifdef NONBLOCKING_RW
				}
			}
#endif
		}
		else if (errno != EINTR)
		{
			log(LOG_ERR,"proxy_main: accept failed: bailing out");
			return 1;
		}
	}
	return 1;
}

