#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <rw.h>
#include <logging.h>
#include <buffers.h>
#include <hdr_parser.h>
#include <hdr_modify.h>

char *read_headers(int fd, char **extra_read, int *extra_len, int *reply_fd)
{
	// Here we assume that 4096 bytes will be able to contain every header
	char buf[4096];
	char *ptr;
	char *ret;
	int nread=0;

	/* Read header */
	do
	{
		int nr;

		if ( (nr = _read(fd,&buf[nread],(sizeof(buf)-nread)-1)) <= 0 )
		{
#ifdef NONBLOCKING_RW
			log(LOG_WARNING,"read_header: read error "
			                "(server connection closed or timed out?)");
#else
			log(LOG_WARNING,"read_header: read error "
			                "(server closed connection?)");
#endif
			return NULL;
		}

		nread += nr;
		buf[nread]='\0';
	}
	while ( ((ptr=get_http_doc_start(buf)) == NULL) && (nread < sizeof(buf)) );

	if ( (ret=malloc(ptr-buf+1)) == NULL )
	{
		log(LOG_ERR,"read_header: Not enough memory");
		return NULL;
	}

	strncpy(ret,buf,ptr-buf);
	ret[ptr-buf]='\0';

	if (!modify_header_except_TE(&ret))
	{
		log(LOG_ERR,"read_header: Not enough memory");
		free(ret);
		return NULL;
	}

	// Reply: header
	if (*reply_fd != -1)
	{
		if (_writen(*reply_fd,ret,strlen(ret)) != strlen(ret))
		{
			*reply_fd = -1;
		}
	}

	if (extra_read != NULL)
	{
		/* Keep the extra_read buffer */
		*extra_len=(&buf[nread]-ptr);

		// We allocate 20 extra bytes here because we could need that
		// later, in the chunk_read_size
		if ( (*extra_read=malloc(*extra_len+20)) == NULL )
		{
			log(LOG_ERR,"read_header: Not enough memory");
			free(ret);
			return NULL;
		}
		memcpy(*extra_read,ptr,*extra_len);
	}

	// Reply: extra_read
	if (*reply_fd != -1)
	{
		if (_writen(*reply_fd,*extra_read,*extra_len) != *extra_len)
		{
			*reply_fd = -1;
		}
	}

	return ret;
}

char *read_sized(int fd, char *firstbuf, int first_buf_size, int size,
                 int *reply_fd)
{
	int nread;
	char *ret;
	char *ptr;

	if ( (ret=malloc(size)) == NULL )
	{
		log(LOG_ERR,"read_sized: Not enough memory");
		return NULL;
	}
	memcpy(ret,firstbuf,MIN(first_buf_size,size));
	nread=MIN(first_buf_size,size);
	ptr=ret+nread;

	if (_readn_reply(fd,ptr,size-nread,reply_fd) < (size-nread))
	{
		log (LOG_ERR, "read_sized: error reading from socket");
		free(ret);
		return NULL;
	}

	return ret;
}

int chunk_read_size(int fd, char *first_buf, int *first_buf_size, int *reply_fd)
{
	char *ptr;
	int nread;
	int len;

	// We must be sure that first_buf is '\0'-terminated
	*(first_buf + *first_buf_size)='\0';

	do
	{
		ptr=first_buf;

		// Goes forward trying to reach an endline or a '\0'
		while ( (*ptr != '\n') && (*ptr != '\r') && (*ptr != '\0') )
			ptr++;

		// If *ptr=='\0' then the message is cut before the end of the
		// chunk-size, so we must read again from the socket
		if (*ptr=='\0')
		{
			/* Here we assume that first_buf is part of a yet allocated array
			 * which will surely be able to contain the chunk-size */
			nread = _read(fd,ptr,10);
			if (nread<=0)
			{
				log(LOG_ERR,"read_chunked: Connection error");
				return -1;
			}

			// Reply
			if (*reply_fd != -1)
			{
				if (_writen(*reply_fd,ptr,nread) != nread)
				{
					*reply_fd = -1;
				}
			}

			*(ptr+nread)='\0';
			*first_buf_size += nread;
		}
		len=strtol(first_buf,&ptr,16);
	}
	while (*ptr=='\0');

	return len;
}

int chunk_read_headers(int fd, char *first_buf, char **hdr, int *reply_fd)
{
	char *ptr;
	char buf[1024];
	int nread=0;

	memcpy(buf,first_buf,sizeof(buf));

	while ( (ptr=get_http_doc_start(buf)) == NULL )
	{
		int nr;
		if ( (nr = _read(fd,&buf[nread],(sizeof(buf)-nread)-1)) <= 0 )
		{
			log(LOG_WARNING,"chunk_read_headers: read error (%d)",nr);
			return 0;
		}

		// Reply
		if (*reply_fd != -1)
		{
			if (_writen(*reply_fd,&buf[nread],nr) != nr)
			{
				*reply_fd=-1;
			}
		}

		nread += nr;
		buf[nread]='\0';
	}

	if ( (ptr-buf) > 3 ) // this is true if we read at least one line
	{
		*(ptr-4)='\0';
		return add_header(hdr,buf,0);
	}
	return 1;
}

char *read_chunked(int fd, char *first_buf, int first_buf_size, char **headers,
                   int *reply_fd)
{
	char cl_hdr[25];
	char *chunk=NULL;
	char *ret=NULL;
	char *ptr;
	int len;
	int totlen=0;

	if ( (len = chunk_read_size(fd,first_buf,&first_buf_size,reply_fd)) < 0 )
	{
		return NULL;
	}

	ptr = first_buf;

	/* Advances until the beginning of the next line or the end of the
	 * buffer - whichever comes first */
	while ( (ptr != (first_buf + first_buf_size) ) && (*((++ptr)-1) != '\n') );

	while (len>0)
	{
		int ptr_len=(first_buf + first_buf_size) - ptr;

		if ( (chunk=read_sized(fd,ptr,ptr_len,len+2,reply_fd)) == NULL )
		{
			free(ret);
			return NULL;
		}

		/* Attach chunk at the rest of the data */
		if ( (bufncat(&ret,totlen,chunk,len)) == NULL )
		{
			log(LOG_ERR,"read_chunked: Not enough memory");
			free(chunk);
			free(ret);
			return NULL;
		}
		free(chunk);

		totlen += len;

		// Delete readen data from first_buf
		if (len+2<ptr_len)
		{
			ptr += len+2;
			ptr_len -= len+2;
			memmove(first_buf,ptr,ptr_len);
		}
		else
			ptr_len=0;

		// Read new block size
		if ( (len = chunk_read_size(fd,first_buf,&ptr_len,reply_fd)) < 0 )
		{
			free(chunk);
			free(ret);
			return NULL;
		}
		first_buf_size=ptr_len;

		ptr=first_buf;
		/* Advances until the beginning of the next line or the end of the
		 * buffer - whichever comes first */
		while ( (ptr != (first_buf+first_buf_size) ) && (*((++ptr)-1)!='\n') )
			;
	}

	// Delete Transfer-encoding: chunked from the headers
	if (!delete_header(headers,"Transfer-Encoding"))
	{
		log(LOG_WARNING,"read_chunked: cannot delete Transfer-Encoding header");
	}

	// Add Content-Length to the headers
	snprintf(cl_hdr,sizeof(cl_hdr),"Content-Length: %d",totlen);
	if (!add_header(headers,cl_hdr,0))
	{
		log(LOG_WARNING,"read_chunked: cannot insert Content-Length header");
	}

	// Read extra headers
	if (!chunk_read_headers(fd,ptr,headers,reply_fd))
	{
		log(LOG_WARNING,"read_chunked: cannot read extra headers");
	}

	return ret;
}

// On successful exit first_buf_size holds the length of the readen document
char *read_connclose(int fd,char *first_buf,int *first_buf_size,int *reply_fd)
{
	char buf[1024];
	char *ret=NULL;
	int nr;
	int totread=0;

	if ( (first_buf != NULL) && (*first_buf_size > 0) )
	{
		if ( (ret = malloc(*first_buf_size)) == NULL )
		{
			log(LOG_ERR,"read_connclose: Not enough memory");
			return NULL;
		}
		memcpy(ret,first_buf,*first_buf_size);
	}

	do
	{
		nr=_read(fd,buf,sizeof(buf));

		// Reply
		if (*reply_fd != -1)
		{
			if (_writen(*reply_fd,buf,nr) < nr)
			{
				*reply_fd = -1;
			}
		}

		if (bufncat(&ret,totread,buf,nr) == NULL)
		{
			log(LOG_ERR,"read_connclose: Not enough memory");
			free(ret);
			return NULL;
		}
		totread += nr;
	}
	while (nr>0);

	*first_buf_size = totread;
	return ret;
}

