#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <hdr_parser.h>
#include <buffers.h>
#include <logging.h>

// both buf and field must be \0 terminated, parsing is case insensitive
char *parse_http_hdr(const char *field, const char *buf)
{
	char *field_safe;
	char *needle;

	if ( (field_safe=malloc(strlen(field)+2)) == NULL )
	{
		log(LOG_ERR,"parse_http_header: Not enough memory");
		return NULL;
	}
	strcpy(field_safe,"\n");
	strcat(field_safe,field);
	needle=strstrcase(buf,field_safe);
	free(field_safe);

	if (needle)
	{
		char *ret=NULL;
		int totlen=0;

		needle += strlen(field)+1;

		do
		{
			int linelen;
			char *tmp;

			while ( (*needle==':') || (is_sptab(*needle)) ) needle++;

			totlen += linelen = strcspn(needle,"\n\r");

			if ( (tmp=malloc(totlen+1)) == NULL )
			{
				log(LOG_ERR,"parse_http_header: Not enough memory");
				free(ret);
				return NULL;
			}

			if (ret != NULL)
			{
				strcpy(tmp,ret);
				free(ret);
			}
			else
			{
				tmp[0]='\0';
			}

			strncat(tmp,needle,linelen);
			tmp[totlen]='\0';

			needle += linelen;
			while ( (*needle=='\n') || (*needle=='\r') ) needle++;

			ret=tmp;
		}
		while (is_sptab(*needle));

		return ret;
	}
	else
		return NULL;
}

char *get_http_doc_start(char *buf)
{
	if ( (strncmp(buf,"\r\n",2)) == 0 )
		return (buf+2);
	while ( ( strncmp(buf,"\r\n\r\n",4) != 0 ) && (*buf != '\0') )
		buf++;
	return (*buf=='\0')?NULL:(buf+4);
}

char *get_requested_host(char *request)
{
	char *host;
	char *ptr;
	char *tmp;

	ptr=request;

	// Advances until requested url
	while ((*ptr!=' ') && (*ptr!='\0'))
		ptr++;
	while (is_sptab(*ptr)) ptr++;

	if ( (strncasecmp(ptr,"http://",7)) != 0 )
	{
		if ((host=parse_http_hdr("Host",request)) == NULL)
		{
			return NULL;
		}

		ptr = host + strcspn(host,":");
		*ptr = '\0';
	}
	else
	{
		ptr+=7;
		tmp=ptr;
		while ((!is_sptab(*tmp)) && (*tmp!='/') && (*tmp!=':') &&
		       (*tmp!='*') && (*tmp!='\0'))
			tmp++;

		if ( (host=malloc((tmp-ptr)+1)) == NULL )
		{
			log(LOG_ERR,"get_host: not enough memory");
			return NULL;
		}
		strncpy(host,ptr,tmp-ptr);
		host[tmp-ptr]='\0';
	}

	return host;
}

int get_requested_port(char *request)
{
	int ret=0;
	char *host;
	char *ptr;

	ptr=request;

	// Advances until requested url
	while ((*ptr!=' ') && (*ptr!='\0'))
		ptr++;
	while (is_sptab(*ptr)) ptr++;

	if ( (strncasecmp(ptr,"http://",7)) != 0 )
	{
		if ((host=parse_http_hdr("Host",request)) == NULL)
		{
			return 0;
		}

		ptr = host + strcspn(host,":");
		if (*ptr != 0)
		{
			ptr++;
			ret = atoi(ptr);
		}
		else
		{
			ret = 0;
		}
		free(host);
	}

	if (ret == 0)
	{
		ptr+=7;

		while ((*ptr != ':') && (!is_sptab(*ptr)) && (*ptr != '\r') &&
		       (*ptr != '\n') && (*ptr != '\0'))
			ptr++;
		if (*ptr == ':')
		{
			ptr++;
			ret = atoi(ptr);
		}
		else
		{
			ret = 0;
		}
	}

	if (ret == 0)
	{
		ret = 80;
	}

	return ret;
}

char *get_requested_path(char *request)
{
	char *path;
	char *ptr;
	char *tmp;

	ptr=request;

	// Advances until requested url
	while ((*ptr!=' ') && (*ptr!='\0'))
		ptr++;
	while (is_sptab(*ptr)) ptr++;

	if ( (strncasecmp(ptr,"http://",7)) == 0 )
	{
		ptr += 7;
		while ( (*ptr != '/') && (*ptr != '\r') && (*ptr != '\n') &&
		        (*ptr != '\0') )
			ptr++;
	}

	tmp=ptr;
	while ((*tmp!=' ') && (*tmp!='\r') && (*tmp!='\n') && (*tmp!='\0'))
		tmp++;

	if (tmp==ptr)
	{
		if ( (path=malloc(2)) == NULL )
		{
			log(LOG_ERR,"get_requested_path: not enough memory");
			return NULL;
		}
		path[0]='/';
		path[1]='\0';
	}
	else
	{
		if ( (path=malloc(tmp-ptr+1)) == NULL )
		{
			log(LOG_ERR,"get_requested_path: not enough memory");
			return NULL;
		}
		strncpy(path,ptr,tmp-ptr);
		path[tmp-ptr]='\0';
	}

	return path;
}

char *get_requested_url(char *request)
{
	char *url;
	char *ptr;
	char *tmp;
	char *host=NULL;

	ptr=request;

	// Advances until requested url
	while ((*ptr!=' ') && (*ptr!='\0'))
		ptr++;
	while (is_sptab(*ptr)) ptr++;

	if ( (strncasecmp(ptr,"http://",7)) != 0 )
	{
		if ((host=parse_http_hdr("Host",request)) == NULL)
		{
			return NULL;
		}
	}

	tmp=ptr;
	while ((*tmp!=' ') && (*tmp!='\0') && (*tmp!='\r') && (*tmp!='\n'))
		tmp++;

	if ( (url=malloc( (host==NULL) ? ((tmp-ptr)+1) :
	                                 ((tmp-ptr)+strlen(host)+8)) ) == NULL )
	{
		log(LOG_ERR,"get_url: not enough memory");
		return NULL;
	}

	if (host==NULL)
	{
		strncpy(url,ptr,tmp-ptr);
		url[tmp-ptr]='\0';
	}
	else
	{
		strcpy(url,"http://");
		strcat(url,host);
		strncat(url,ptr,tmp-ptr);
		url[(tmp-ptr)+strlen(host)+7]='\0';
	}

	return url;
}

int get_request_type(char *request)
{
	int ret=REQ_UNKNOWN;

	if (strncasecmp(request,"GET ",4)==0)
	{
		ret=REQ_GET;
	}
	if (strncasecmp(request,"HEAD ",5)==0)
	{
		ret=REQ_HEAD;
	}
	if (strncasecmp(request,"POST ",5)==0)
	{
		ret=REQ_POST;
	}
	if (strncasecmp(request,"PUT ",4)==0)
	{
		ret=REQ_PUT;
	}
	if (strncasecmp(request,"OPTIONS ",8)==0)
	{
		ret=REQ_OPTIONS;
	}
	if (strncasecmp(request,"DELETE ",7)==0)
	{
		ret=REQ_DELETE;
	}
	if (strncasecmp(request,"TRACE ",6)==0)
	{
		ret=REQ_TRACE;
	}

	return ret;
}

int get_reply_type(char *header)
{
	if (strncasecmp(header,"HTTP/1.1",8) != 0)
	{
		log(LOG_WARNING,"get_reply_type: server is NOT speaking http 1.1");
	}
	// Jump over HTTP/1.1
	while ( (*header != '\0') && (*header != ' ') )
		header++;

	// Reach error code
	while ( (*header != '\0') && (is_sptab(*header)) )
		header++;

	return (atoi(header));
}

int there_is_data(char *request, char *header)
{
	int r_type;
	int type;

	r_type=(request==NULL)?REQ_GET:get_request_type(request);
	type=get_reply_type(header);

	if (r_type == REQ_HEAD)
	{
		return 0;
	}

	// OPTIONS replies hold data only if reply is not an error
	if (r_type == REQ_OPTIONS)
	{
		if ( (type % 100) == 2 )
			return 0;
	}

	// 204 No Content
	// 205 Reset Content
	// 304 Not Modified

	if ( (type == 204) || (type == 205) || (type == 304) )
	{
		return 0;
	}

	return 1;
}

