#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>

#ifdef NONBLOCKING_RW

#include <sys/time.h>

// use timeout passed as value of NONBLOCKING_RW
#define TIMEOUT NONBLOCKING_RW

// Simply calls read, checking if errno is EINTR,
// and eventually repeating the call
int _read(int fd, char *buf, unsigned int n)
{
	struct timeval timeout;
	fd_set rset,wset,eset;
	int nread, cycle = 1;

	timeout.tv_sec=TIMEOUT;
	timeout.tv_usec=0;

	while ( cycle )
	{
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_ZERO(&eset);

		FD_SET(fd,&rset);

		if (select(fd+1,&rset,&wset,&eset,&timeout) < 0)
		{
			return -errno;
		}

		if ( (nread = read(fd, buf, n)) >= 0 )
			cycle = 0;
		else if ( errno != EINTR )
			return -errno;
	}
	return nread;
}

// Simply calls write, checking if errno is EINTR,
// and eventually repeating the call
int _write(int fd, const char *buf, unsigned int n)
{
	struct timeval timeout;
	fd_set rset,wset,eset;
	int nwritten, cycle = 1;

	timeout.tv_sec=TIMEOUT;
	timeout.tv_usec=0;

	while ( cycle )
	{
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_ZERO(&eset);

		FD_SET(fd,&wset);

		if (select(fd+1,&rset,&wset,&eset,&timeout) < 0)
		{
			return -errno;
		}

		if ( (nwritten = write(fd, buf, n)) > 0 )
			cycle = 0;
		else if ( errno != EINTR )
			return -errno;
	}
	return nwritten;
}

// Reads _exactly_ n characters, blocking the calling process
int _readn(int fd, void *buf, unsigned int n)
{
	struct timeval timeout;
	fd_set rset,wset,eset;
	int nread;
	char *ptr;

	ptr = buf;

	while (n>0)
	{
		timeout.tv_sec=TIMEOUT;
		timeout.tv_usec=0;

		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_ZERO(&eset);

		FD_SET(fd,&rset);

		if (select(fd+1,&rset,&wset,&eset,&timeout) < 0)
		{
			return -errno;
		}

		if ((nread = read(fd,ptr,n)) < 0)
		{
			if (errno==EINTR)
				nread=0;
			else
				return -1;
		}
		else if (nread==0)
			break;
		n -= nread;
		ptr += nread;
	}
	return (ptr - (char *)buf);
}

// Writes _exactly_ n characters, blocking the calling process
int _writen(int fd, const void *buf, unsigned int n)
{
	struct timeval timeout;
	fd_set rset,wset,eset;
	int nwritten;
	char *ptr;

	ptr = (char *)buf;

	while (n>0)
	{
		timeout.tv_sec=TIMEOUT;
		timeout.tv_usec=0;

		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_ZERO(&eset);

		FD_SET(fd,&wset);

		if (select(fd+1,&rset,&wset,&eset,&timeout) < 0)
		{
			return -errno;
		}

		if ((nwritten=write(fd,ptr,n)) <= 0)
		{
			if (errno==EINTR)
				nwritten=0;
			else
				return -1;
		}
		n -= nwritten;
		ptr += nwritten;
	}
	return (ptr - (char *)buf);
}

// Reads _exactly_ n characters, blocking the calling process
// (if doesn't time out), replying everything on *reply_fd (for streaming)
int _readn_reply(int fd, void *buf, unsigned int n, int *reply_fd)
{
	struct timeval timeout;
	fd_set rset,wset,eset;
	int nread;
	char *ptr;

	ptr = buf;

	while (n>0)
	{
		timeout.tv_sec=TIMEOUT;
		timeout.tv_usec=0;

		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_ZERO(&eset);

		FD_SET(fd,&rset);

		if (select(fd+1,&rset,&wset,&eset,&timeout) < 0)
		{
			return -errno;
		}

		if ((nread = read(fd,ptr,n)) < 0)
		{
			if (errno==EINTR)
				nread=0;
			else
				return -1;
		}
		else if (nread==0)
			break;

		if (*reply_fd != -1)
		{
			if (_writen(*reply_fd,ptr,nread) != nread)
			{
				*reply_fd = -1;
			}
		}
		n -= nread;
		ptr += nread;
	}
	return (ptr - (char *)buf);
}

#else

// Simply calls read, checking if errno is EINTR,
// and eventually repeating the call
int _read(int fd, char *buf, unsigned int n)
{
	int nread, cycle = 1;
	while ( cycle )
	{
		if ( (nread = read(fd, buf, n)) >= 0 )
			cycle = 0;
		else if ( errno != EINTR )
			return -errno;
	}
	return nread;
}

// Simply calls write, checking if errno is EINTR,
// and eventually repeating the call
int _write(int fd, const char *buf, unsigned int n)
{
	int nwritten, cycle = 1;
	while ( cycle )
	{
		if ( (nwritten = write(fd, buf, n)) > 0 )
			cycle = 0;
		else if ( errno != EINTR )
			return -errno;
	}
	return nwritten;
}

// Reads _exactly_ n characters, blocking the calling process
int _readn(int fd, void *buf, unsigned int n)
{
	int nread;
	char *ptr;

	ptr = buf;

	while (n>0)
	{
		if ((nread = read(fd,ptr,n)) < 0)
		{
			if (errno==EINTR)
				nread=0;
			else
				return -1;
		}
		else if (nread==0)
			break;
		n -= nread;
		ptr += nread;
	}
	return (ptr - (char *)buf);
}

// Writes _exactly_ n characters, blocking the calling process
int _writen(int fd, const void *buf, unsigned int n)
{
	int nwritten;
	char *ptr;

	ptr = (char *)buf;

	while (n>0)
	{
		if ((nwritten=write(fd,ptr,n)) <= 0)
		{
			if (errno==EINTR)
				nwritten=0;
			else
				return -1;
		}
		n -= nwritten;
		ptr += nwritten;
	}
	return (ptr - (char *)buf);
}

// Reads _exactly_ n characters, blocking the calling process,
// replying everything on *reply_fd (for streaming)
int _readn_reply(int fd, void *buf, unsigned int n, int *reply_fd)
{
	int nread=0;
	char *ptr;

	ptr = buf;

	while (n>0)
	{
		if ((nread = read(fd,ptr,n)) < 0)
		{
			if (errno==EINTR)
				nread=0;
			else
				return -1;
		}
		else if (nread==0)
			break;

		if (*reply_fd != -1)
		{
			if (_writen(*reply_fd,ptr,nread) != nread)
			{
				*reply_fd = -1;
			}
		}
		n -= nread;
		ptr += nread;
	}
	return (ptr - (char *)buf);
}

#endif

